#Adapted from calculate_mpdockq.py (https://github.com/KosinskiLab/AlphaPulldown/blob/main/alphapulldown/analysis_pipeline/alpha_analysis_jax0.4.def)

import numpy as np
import pandas as pd
from collections import defaultdict
import math

################FUNCTIONS#################
def parse_atm_record(line):
    '''Get the atm record
    '''
    record = defaultdict()
    record['name'] = line[0:6].strip()
    record['atm_no'] = int(line[6:11].strip())
    record['atm_name'] = line[12:16].strip()
    record['atm_alt'] = line[17]
    record['res_name'] = line[17:20].strip()
    record['chain'] = line[21]
    record['res_no'] = int(line[22:26].strip())
    record['insert'] = line[26].strip()
    record['resid'] = line[22:29]
    record['x'] = float(line[30:38])
    record['y'] = float(line[38:46])
    record['z'] = float(line[46:54])
    record['occ'] = float(line[54:60])
    record['B'] = float(line[60:66])

    return record

def read_pdb(pdbfile):
    '''Read a pdb file per chain
    '''
    pdb_chains = {}
    chain_coords = {}
    chain_CA_inds = {}
    chain_CB_inds = {}

    with open(pdbfile) as file:
        for line in file:
            if 'ATOM' in line:
                record = parse_atm_record(line)
                if record['chain'] in [*pdb_chains.keys()]:
                    pdb_chains[record['chain']].append(line)
                    chain_coords[record['chain']].append([record['x'],record['y'],record['z']])
                    coord_ind+=1
                    if record['atm_name']=='CA':
                        chain_CA_inds[record['chain']].append(coord_ind)
                    if record['atm_name']=='CB' or (record['atm_name']=='CA' and record['res_name']=='GLY'):
                        chain_CB_inds[record['chain']].append(coord_ind)


                else:
                    pdb_chains[record['chain']] = [line]
                    chain_coords[record['chain']]= [[record['x'],record['y'],record['z']]]
                    chain_CA_inds[record['chain']]= []
                    chain_CB_inds[record['chain']]= []
                    #Reset coord ind
                    coord_ind = 0


    return pdb_chains, chain_coords, chain_CA_inds, chain_CB_inds

def read_plddt(best_plddt, chain_CA_inds):
    '''Get the plDDT for each chain
    '''
    chain_names = chain_CA_inds.keys()
    chain_lengths = dict()
    for name in chain_names:
        curr_len = len(chain_CA_inds[name])
        chain_lengths[name] = curr_len
    
    plddt_per_chain = dict()
    curr_len = 0
    for k,v in chain_lengths.items():
        curr_plddt = best_plddt[curr_len:curr_len+v]
        plddt_per_chain[k] = curr_plddt
        curr_len += v 
    return plddt_per_chain

def score_complex(path_coords, path_CB_inds, path_plddt):
    '''
    Score all interfaces in the current complex

    Modified from the score_complex() function in MoLPC repo: 
    https://gitlab.com/patrickbryant1/molpc/-/blob/main/src/complex_assembly/score_entire_complex.py#L106-154
    '''

    chains = [*path_coords.keys()]
    chain_inds = np.arange(len(chains))
    complex_score = 0
    #Get interfaces per chain
    for i in chain_inds:
        chain_i = chains[i]
        chain_coords = np.array(path_coords[chain_i])
        chain_CB_inds = path_CB_inds[chain_i]
        l1 = len(chain_CB_inds)
        chain_CB_coords = chain_coords[chain_CB_inds]
        chain_plddt = path_plddt[chain_i]
 
        for int_i in np.setdiff1d(chain_inds, i):
            int_chain = chains[int_i]
            int_chain_CB_coords = np.array(path_coords[int_chain])[path_CB_inds[int_chain]]
            int_chain_plddt = path_plddt[int_chain]
            #Calc 2-norm
            mat = np.append(chain_CB_coords,int_chain_CB_coords,axis=0)
            a_min_b = mat[:,np.newaxis,:] -mat[np.newaxis,:,:]
            dists = np.sqrt(np.sum(a_min_b.T ** 2, axis=0)).T
            contact_dists = dists[:l1,l1:]
            contacts = np.argwhere(contact_dists<=8)
            #The first axis contains the contacts from chain 1
            #The second the contacts from chain 2
            if contacts.shape[0]>0:
                av_if_plDDT = np.concatenate((chain_plddt[contacts[:,0]], int_chain_plddt[contacts[:,1]])).mean()
                complex_score += np.log10(contacts.shape[0]+1)*av_if_plDDT

    return complex_score, len(chains)

def calculate_mpDockQ(complex_score):
    """
    A function that returns a complex's mpDockQ score after 
    calculating complex_score
    """
    L = 0.827
    x_0 = 261.398
    k = 0.036
    b = 0.221
    return L/(1+math.exp(-1*k*(complex_score-x_0))) + b

def read_pdb_pdockq(pdbfile):
    '''Read a pdb file predicted with AF and rewritten to conatin all chains
    Adepted from FoldDock repo:
    https://gitlab.com/ElofssonLab/FoldDock/-/blob/main/src/pdockq.py#L34-59
    '''

    chain_coords, chain_plddt = {}, {}
    with open(pdbfile, 'r') as file:
        for line in file:
            if not line.startswith('ATOM'):
                continue
            record = parse_atm_record(line)
            #Get CB - CA for GLY
            if record['atm_name']=='CB' or (record['atm_name']=='CA' and record['res_name']=='GLY'):
                if record['chain'] in [*chain_coords.keys()]:
                    chain_coords[record['chain']].append([record['x'],record['y'],record['z']])
                    chain_plddt[record['chain']].append(record['B'])
                else:
                    chain_coords[record['chain']] = [[record['x'],record['y'],record['z']]]
                    chain_plddt[record['chain']] = [record['B']]


    #Convert to arrays
    for chain in chain_coords:
        chain_coords[chain] = np.array(chain_coords[chain])
        chain_plddt[chain] = np.array(chain_plddt[chain])

    return chain_coords, chain_plddt


def calc_pdockq(chain_coords, chain_plddt, t):
    '''Calculate the pDockQ scores
    pdockQ = L / (1 + np.exp(-k*(x-x0)))+b
    L= 0.724 x0= 152.611 k= 0.052 and b= 0.018

    Modified from the calc_pdockq() from FoldDock repo: 
    https://gitlab.com/ElofssonLab/FoldDock/-/blob/main/src/pdockq.py#L62
    '''

    #Get coords and plddt per chain
    ch1, ch2 = [*chain_coords.keys()]
    coords1, coords2 = chain_coords[ch1], chain_coords[ch2]
    plddt1, plddt2 = chain_plddt[ch1], chain_plddt[ch2]

    #Calc 2-norm
    mat = np.append(coords1, coords2,axis=0)
    a_min_b = mat[:,np.newaxis,:] -mat[np.newaxis,:,:]
    dists = np.sqrt(np.sum(a_min_b.T ** 2, axis=0)).T
    l1 = len(coords1)
    contact_dists = dists[:l1,l1:] #upper triangular --> first dim = chain 1
    contacts = np.argwhere(contact_dists<=t)

    if contacts.shape[0]<1:
        pdockq=0
    else:
        #Get the average interface plDDT
        avg_if_plddt = np.average(np.concatenate([plddt1[np.unique(contacts[:,0])], plddt2[np.unique(contacts[:,1])]]))
        #Get the number of interface contacts
        n_if_contacts = contacts.shape[0]
        x = avg_if_plddt*np.log10(n_if_contacts)
        pdockq = 0.724 / (1 + np.exp(-0.052*(x-152.611)))+0.018

    return pdockq


import os
import pickle
import numpy as np
import matplotlib.pyplot as plt
import csv
import logging
import subprocess
import multiprocessing
import pynvml
import copy
import sys
import signal
import glob
import pandas as pd
from pathlib import Path
from datetime import datetime
from Bio import SeqIO
import time
import queue
from concurrent.futures import ThreadPoolExecutor, as_completed
import math


# Configure global logger
logging.basicConfig(
    filename="./log_file/HInt.log",  # Log file name
    level=logging.INFO,  # Log level
    format="%(asctime)s - %(levelname)s - %(message)s"  # Log format
)

# Use the logger configuration from HInt.py
logger = logging.getLogger()



def Define_informations() :
    """
    Parse the HInt.txt configuration file and collect all user-defined parameters into a single dictionary.

    - Reads key : value pairs from HInt.txt (comments starting with # are ignored)
    - Verifies the presence of mandatory parameters
    - Assigns default values to optional or empty parameters
    - Normalizes paths (removes trailing '/')
    - Parses interaction definitions (including multimers and interaction regions)
    - Validates biological constraints (DeepLoc, homo-oligomer, number of baits)

    Returns:
    -------
    Informations_dict : dict
        Dictionary containing all validated and formatted configuration parameters required by HInt.

    """
    logger.info("Defining informations")
    Informations_dict = dict()
    list_inf = ["Signal_peptide", "Homo-oligomer", "Interact_with", "Organism", "DeepLoc", "Regions", "Multimer_bait", "AlphaFold", "Max_protein_lenght", "Min_protein_lenght", "Path_Uniprot_ID", "Path_AlphaFold_Data", "Path_Pickle_Feature", "Path_Singularity_Image", "Path_MMseqs2_Data"]
    with open("HInt.txt", "r") as file :
        for lines in file :
            if ":" in lines :
                info = lines.split("#")[0]
                informations_name = info.split(":")[0].strip().strip("\n")
                informations = info.split(":")[1].strip().strip("\n")
                Informations_dict[informations_name] = informations
    for info in list_inf :
        if info not in Informations_dict.keys() : #if settings file is not authentic
            if info in ["Interact_with", "Organism","Path_Uniprot_ID", "Path_AlphaFold_Data", "Path_Pickle_Feature"] :
                raise ValueError(f"HInt.txt file is compromised, verify the file. {info} is missing")
            elif info in ["Signal_peptide","Homo-oligomer","Path_MMseqs2_Data","Regions","Multimer_bait","DeepLoc","AlphaFold","Max_protein_lenght","Min_protein_lenght"] :
                Informations_dict[info] = ""

    ### Normalize all configuration values
    for informations_key in Informations_dict.keys() : #verify all informations and set default value
        if type(Informations_dict[informations_key]) is str and Informations_dict[informations_key].endswith("/") : #avoid error in path
            Informations_dict[informations_key] = Informations_dict[informations_key][:-1]
        if len(Informations_dict[informations_key]) == 0 :
            logger.info(f'{informations_key} is empty')
            if informations_key == "Path_ccp4" :
                if os.path.isdir("/opt/xtal/ccp4-9") == True :
                    logger.info("Set ccp4 path by default on /opt/xtal/ccp4-9")
                    Informations_dict[informations_key] = "/opt/xtal/ccp4-9"
                else :
                    raise ValueError("Path_ccp4 is empty and ccp4 is not found in /opt/xtal/ccp4-9. You need to install ccp4 or set the path of ccp4 in HInt.txt")
            elif informations_key == "Path_AlphaFold_Data" :
                if os.path.isdir("./alphadata") == True :
                    logger.info("Set AlphaFold data by default on ./alphadata")
                    Informations_dict[informations_key] = "./alphadata"
                else :
                    raise ValueError("Path_AlphaFold_Data is empty and AlphaFold data is not found in ./alphadata. You need to download AlphaFold data or set the path of AlphaFold data in HInt.txt")
            elif informations_key == "Path_Pickle_Feature" :
                logger.info("Set pickle feature path by default on ./feature")
                Informations_dict[informations_key] = "./feature"
            elif informations_key == "Path_MMseqs2_Data" :
                logger.info("/!\ local MMseqs2 GPU will not be used")
            elif informations_key == "Signal_peptide" :
                Informations_dict[informations_key] = "None"
            elif informations_key == "Homo-oligomer" :
                Informations_dict[informations_key] = "1"
            elif informations_key == "DeepLoc" :
                Informations_dict[informations_key] = "None"
            elif informations_key == "AlphaFold" :
                Informations_dict[informations_key] = "2"
                logger.info("Set AlphaFold version by default on AlphaFold2")
            elif informations_key == "Min_protein_lenght" :
                Informations_dict[informations_key] = "20"
                logger.info("Minimum lenght for prey protein set by default 20 to AA")
            elif informations_key == "Max_protein_lenght" :
                Informations_dict[informations_key] = ""
        if len(Informations_dict[informations_key]) == 1 :
            if informations_key == "Path_AlphaFold_Data" :
                if os.path.isdir(Informations_dict[informations_key]) == False :
                    raise ValueError(f"Path set for AlphaFold database doesn't exist : {Informations_dict[informations_key]}")
            elif informations_key == "Path_MMseqs2_Data" :
                if os.path.isdir(Informations_dict[informations_key]) == False :
                    raise ValueError (f"Path set for MMseq database doesn't exist : {Informations_dict[informations_key]}")
            elif informations_key == "Path_ccp4" :
                if os.path.isdir(Informations_dict[informations_key]) == False :
                    raise ValueError (f"Path set for ccp4 doesn't exist : {Informations_dict[informations_key]}")
        ### Parse protein-protein interaction definitions
        # Multiple baits
        # Multimeric baits
        # Region-specific interactions
        if informations_key == "Interact_with" :
            regions_dict = dict()
            new_baits_list = list()
            if "[" in Informations_dict["Interact_with"] : #create multimer for bait
                list_complex = [prot.strip("[").strip(",").strip() for prot in Informations_dict["Interact_with"].split("]")]
                if "" in list_complex :
                    list_complex.remove("")
                nbr_baits = len(list_complex)
                list_baits = [prot.strip("[").strip("]") for prot in Informations_dict["Interact_with"].split(",")]
                Informations_dict["Multimer_bait"] = list_complex
            else :
                list_baits = [prot.strip(",").strip() for prot in Informations_dict["Interact_with"].split(",")]
                nbr_baits = len(list_baits)
                Informations_dict["Multimer_bait"] = list_baits
            if nbr_baits > 3 :
                raise ValueError("HInt don't support more than 3 differents baits")
            for prot in list_baits :
                if "-" in prot :
                    name_prot = prot.split("(")[0]
                    new_baits_list.append(name_prot.strip())
                    regions_dict[name_prot.strip()] = prot.split("(")[1].strip(")")
                    for i,multimer in enumerate(Informations_dict["Multimer_bait"]) :
                        if prot in multimer :
                            new_multimer = multimer.replace(prot,name_prot.strip())
                            Informations_dict["Multimer_bait"][i] = new_multimer
                else :
                    if prot.strip() in regions_dict.keys() and regions_dict[prot.strip()] != "0-0" :
                        raise ValueError(f"HInt don't support multiple regions for protein bait : {prot.strip()}")
                    regions_dict[prot.strip()] = "0-0"
                    new_baits_list.append(prot.strip())
            Informations_dict["Regions"] = regions_dict
            Informations_dict["Interact_with"] = new_baits_list

        ### Homo-oligomer check
        if informations_key == "Homo-oligomer" :
            if int(Informations_dict[informations_key]) == False :
                raise ValueError(f"Homo-oligomer is not an integer")
            if Informations_dict[informations_key] == "0" :
                Informations_dict[informations_key] = "1"
        ### DeepLoc check
        if informations_key == "DeepLoc" :
            if Informations_dict["Organism"] == "euk" : #euk
                for value in Informations_dict[informations_key].split(","):
                    if value.strip()  not in ["Cytoplasm", "Nucleus", "Extracellular", "Cell membrane", "Mitochondrion", "Plastid", "Endoplasmic reticulum", "Lysosome/Vacuole", "Golgo apparatus", "Peroxisome","None"] :
                        raise ValueError(f"Incorrect DeepLoc value : {value}")
            else : #other
                for value in Informations_dict[informations_key].split(","):
                    if value.strip()  not in ["Cell wall & surface","Extracellular","Cytoplasmic","Cytoplasmic Membrane","Outer Membrane","Periplasmic","None"] :
                        raise ValueError(f"Incorrect DeepLocPro value : {value}")

    if len(Informations_dict["Signal_peptide"]) == 0 and len(Informations_dict["DeepLoc"]) == 0 and len(Informations_dict["Homo-oligomer"]) == 0 and len(Informations_dict["Interact_with"]) == 0 : #no info
        logger.info("Need information to discriminate the potential homolog/interolog")
        exit()
    return(Informations_dict)

def run_deeploc(file, org, need_DeepLoc, GPU) :
    """
    Run DeepLoc (DeepLoc2 or DeepLocPro) to predict the subcellular localization of proteins requiring DeepLoc annotation.

    - Extracts protein sequences (with signal peptide handling) from a File_proteins object
    - Writes a temporary FASTA file containing only proteins needing DeepLoc prediction
    - Runs the appropriate DeepLoc software depending on the organism type
    - Parses the output CSV file
    - Stores predicted localizations in the result dictionary

    Parameters :
    ----------
    file : object of class File_proteins
    org : str
    need_DeepLoc : list
    GPU : list

    """
    result_dict = file.get_result_dict()
    prot_seq = file.get_proteins_sequence_SP()
    deeploc = file.get_deeploc()
    GPU_str = ""
    for nbr_GPU in GPU :
        GPU_str += nbr_GPU + ","
    GPU_str = GPU_str.strip(",")
    file_name = file.get_file_name().split("/")[-1]
    fasta_file = file_name.replace(".txt","_msa.fasta")


    if os.path.exists("log_file/result_deeploc") == True :
        os.system("rm -r log_file/result_deeploc")
    dp_lines = str()
    for protein in need_DeepLoc :
        dp_lines += ">"+protein+"\n"+prot_seq[protein]+"\n"
    with open(f"log_file/{fasta_file}", "w") as dp_file :
        dp_file.write(dp_lines)

    if org == "euk" :
        logger.info("Start DeepLoc eucaryote")
        software = "deeploc2"
        cmd = f"CUDA_VISIBLE_DEVICES={GPU_str} {software} -f log_file/{fasta_file} -o log_file/result_deeploc -d cuda"
    else :
        logger.info("Start DeepLocPro")
        software = "deeplocpro"
        if org == "gram-" :
            group = "negative"
        if org == "gram+" :
            group = "positive"
        if org == "arch" :
            group = "archaea"
        cmd = f"CUDA_VISIBLE_DEVICES={GPU_str} {software} -f log_file/{fasta_file} -o log_file/result_deeploc -d cuda -g {group}" #deeplocpro don't use all GPU
    os.system(cmd)
    DeepLoc_file = os.listdir(f"log_file/result_deeploc")
    if software == "deeplocpro" :
        min_index = 3
        min_score = 0.1
        index_name = 1
    else :
        min_index = 4
        min_score = 0.4
        index_name = 0
    with open(f"log_file/result_deeploc/{DeepLoc_file[0]}", "r") as DL_file :
        reader = csv.reader(DL_file, delimiter=',')
        for index, line in enumerate(reader) :
            compartment = str()
            if index == 0 :
                first_line = line #save title name
            else :
                protein = line[index_name]
                for index, score in enumerate(line) :
                    if index >= min_index and float(score) > min_score :
                        compartment += first_line[index] + "|"
                compartment = compartment.strip("|")
                
                deeploc[protein] = compartment
                result_dict[protein]["DeepLoc"] = deeploc[protein]
    file.set_result_dict(result_dict)
    file.set_deeploc(deeploc)



def run_SP (file, Informations_dict, need_SP, need_msa) :
    """
    Remove signal peptides from protein sequences using SignalP and generate new FASTA sequences without signal peptides.

    - Runs SignalP on proteins requiring signal peptide (SP) detection
    - Extracts cleavage positions from SignalP output
    - Generates new sequences without signal peptides
    - Updates internal sequence storage
    - Returns the list of proteins still requiring MSA computation
    - Updates the result dictionary with signal peptide presence/absence

    If a protein already has an MSA but does not yet have a sequence without signal peptide, it will be processed here but not returned for MSA.

    Parameters :
    ----------
    file : object of class File_proteins
    Informations_dict : dict
    need_SP : list 
        List of protein identifiers requiring signal peptide detection/removal.
    need_msa : list
        List of protein identifiers requiring MSA generation.

    Returns :
    ----------
    new_need_msa : list
    """
    final_file = str()
    SP_signal = 0
    file_name = file.get_file_name().split("/")[-1]
    fasta_file = file_name.replace(".txt","_msa.fasta")
    output_file = fasta_file.replace(".fasta","")
    new_fasta_dict = file.get_proteins_sequence_no_SP()
    result_dict = file.get_result_dict()
    if_prot_SP = file.get_prot_SP()
    prot_SP_len = dict()

    file.create_fasta_file(True, need_SP)
    cmd = f"signalp -fasta log_file/{fasta_file} -org {Informations_dict['Organism']} -prefix log_file/{output_file}"
    os.system(cmd)

    file_signalp = fasta_file.replace(".fasta","_summary.signalp5")
    with open(f"log_file/{file_signalp}","r") as fh :
        for line in fh :
            new_line = line.split("\t")
            if new_line[1] != "OTHER" and new_line[0][0] != "#" :
                prot_SP_len[new_line[0]] = new_line[len(new_line)-1].split("-")[1].split(".")[0]
                result_dict[new_line[0]]["Signal_peptide"] = "Yes"
                if_prot_SP[new_line[0]] = "Yes"
            elif new_line[1] == "OTHER" and new_line[0][0] != "#" :
                result_dict[new_line[0]]["Signal_peptide"] = "No"
                if_prot_SP[new_line[0]] = "No"
    with open(f"log_file/{fasta_file}", "r") as fa_file :
        for line2 in fa_file :
            new_line2 = line2
            if SP_signal == 0 and line2[0] != ">" :
                new_line2 = line2
                new_fasta_dict[save_key] = line2.strip("\n")
            if int(SP_signal) > 0 :
                new_line2 = line2[int(SP_signal)-1:len(line2)]
                new_fasta_dict[save_key] = line2[int(SP_signal)-1:len(line2)].strip("\n")
                SP_signal = 0
            if line2[0] == ">" :
                save_key = line2[1:len(line2)-1]
                if str(line2[1:len(line2)-1]) in prot_SP_len.keys() :
                    SP_signal = prot_SP_len[line2[1:len(line2)-1]]
            final_file = final_file + new_line2

    new_need_msa = list()
    for prot in need_msa : 
        if os.path.isfile(f"{Informations_dict['Path_Pickle_Feature']}/{prot}.a3m") == False :
            new_need_msa.append(prot)
    file.set_proteins_sequence_no_SP(new_fasta_dict)
    file.set_result_dict(result_dict)
    file.set_prot_SP(if_prot_SP)
    return new_need_msa

def create_feature (file, Informations_dict, GPU, CPU, need_msa, need_pkl) :
    """
    Generate MSAs and AlphaFold feature pickle files for a set of proteins.

    - Searches for precomputed MSAs in the AlphaFold database
    - Generates MSAs using ColabFold/MMseqs2 when missing
    - Uses CPU parallelization and GPU acceleration
    - Updates the lists of proteins requiring MSA or pickle generation

    Parameters :
    ----------
    file : object of class File_proteins
    informations_dict : dict
    GPU : list
    CPU : int
    need_msa : list
    need_pkl : list

    """
    Path_AlphaFold_Data = Informations_dict["Path_AlphaFold_Data"]
    Path_Pickle_Feature = Informations_dict["Path_Pickle_Feature"]
    Path_MMseqs2_Data = Informations_dict["Path_MMseqs2_Data"]
    baits = Informations_dict["Interact_with"]
    prot_no_SP = file.get_proteins_sequence_no_SP()
    prot_SP = file.get_proteins_sequence_SP()
    file_name = file.get_file_name()
    msa_name = file_name.replace(".txt","_msa.fasta")
    pkl_name = file_name.replace(".txt","_pkl.fasta")
    GPU_str = ""
    for nbr_GPU in GPU :
        GPU_str += nbr_GPU + ","
    GPU_str = GPU_str.strip(",")
    #logger.info(f"GPU use : {GPU_str}")
    
    logger.info(f"{len(need_msa)} proteins need MSA")
    logger.info(f"{len(need_pkl)} proteins need pkl files")
    
    if os.path.exists(Path_Pickle_Feature) == False :
        os.system(f"mkdir {Path_Pickle_Feature}")

    #Look for MSA in AlphaFold database
    generated_msa = copy.deepcopy(need_msa)
    logger.info(f"Search MSA in AlphaFold database")

    # ThreadPool to parallelise
    cpu_per_mafft = 2
    max_workers = max(1, CPU // cpu_per_mafft)  # how many mafft in parallel


    start = time.time()

    futures_list = []
    with ThreadPoolExecutor(max_workers=max_workers) as executor : #CPU parallelization
        for protein in generated_msa : #check if prot have an MSA in alphafold database
            l_p = len(protein)
            if l_p >= 5 and l_p <= 10 and "_" not in protein : #if not, is not an UniprotID
                futures_list.append(executor.submit(fetch_trim_mafft,protein,Path_Pickle_Feature,prot_SP,prot_no_SP,))

        for future in as_completed(futures_list) :
            protein, ok = future.result()
            if ok :
                logger.info(f"MSA for {protein} processed")
                need_msa.remove(protein) #msa found
                need_pkl.append(protein)
                    
    logger.info("MSA search in AlphaFold database complete")

    file.create_fasta_file(False, need_msa, need_pkl)
    #Create MSA files with ColabFold mmseq2 GPU accelerated for proteins without MSA
    if len(need_msa) > 10 :
        if len(GPU_str.split(",")) >= 4 : #Due to error by using mmseqGPU with more than 3 GPU 
            GPU_str = GPU_str[:-2]
        cmd = f"CUDA_VISIBLE_DEVICES={GPU_str} colabfold_search ./log_file/{msa_name} {Path_MMseqs2_Data} {Path_Pickle_Feature} --db-load-mode 2 --gpu 1"  #-e 0.1
        os.system(cmd)
        need_pkl.extend(need_msa)
        need_msa = list()

    if len(need_msa) < 1 :
        logger.info("All MSAs have already been generated")

    file.create_fasta_file(False, need_msa, need_pkl)

    #Create MSA files with ColabFold mmseq2 classic pipeline for proteins without MSA, and pickle files for proteins generated with MMseqs2 GPU
    if len(need_msa) > 0 and len(need_msa) <= 10 :
        if os.path.isfile(f"log_file/{msa_name}") == True :
            cmd = ["create_individual_features.py",
            f"--fasta_paths=./log_file/{msa_name}",
            f"--data_dir={Path_AlphaFold_Data}",
            "--save_msa_files=True",
            f"--output_dir={Path_Pickle_Feature}",
            "--max_template_date=2024-05-02",
            "--skip_existing=True",
            "--use_mmseqs2=True",
            "--use_precomputed_msas=True"]
            process = subprocess.Popen(cmd, stderr=subprocess.STDOUT, text=True, bufsize=1, universal_newlines=True)
            stdout, stderr = process.communicate()

    end = time.time()
    elapsed = end - start
    logger.info("Create MSA take "+ str(elapsed/60)+" minutes")

    #Create pkl files for proteins without pkl file (from MSA found in AFdb or error during MSA generation with ColabFold mmseqs2)
    if os.path.isfile(f"log_file/{pkl_name}") == True :
        cmd2 = ["create_individual_features.py",
        f"--fasta_paths=./log_file/{pkl_name}",
        f"--data_dir={Path_AlphaFold_Data}",
        "--save_msa_files=True",
        f"--output_dir={Path_Pickle_Feature}",
        "--max_template_date=2024-05-02",
        "--skip_existing=True",
        "--use_mmseqs2=True",
        "--use_precomputed_msas=True"]
        process = subprocess.Popen(cmd2, stderr=subprocess.STDOUT, text=True, bufsize=1, universal_newlines=True)
        stdout, stderr = process.communicate()
            

def fetch_trim_mafft(protein, Path_Pickle_Feature, prot_SP, prot_no_SP) :
    """
    Retrieve a precomputed MSA from the AlphaFold database (AFdb), trim signal peptides if present, and realign the MSA using MAFFT.

    - Checks whether an MSA exists for the given protein in AFdb
    - Downloads the MSA in A3M format if available
    - Removes signal peptide residues from all sequences when necessary
    - Filters out excessively short or invalid sequences after trimming
    - Realigns the trimmed MSA using MAFFT
    - Reformats the alignment back to A3M format

    Parameters :
    ----------
    protein : str
    Path_Pickle_Feature : str
    prot_SP : dict
    prot_no_SP : dict

    Returns :
    ----------
    protein : str
    bool
        True if an MSA was successfully retrieved and processed.
        False if no MSA was found in AFdb.
    """
    url = f"https://alphafold.ebi.ac.uk/files/msa/AF-{protein}-F1-msa_v6.a3m"
    msa_in = f"{Path_Pickle_Feature}/{protein}.a3m"
    aln_out = f"{Path_Pickle_Feature}/{protein}.aln"


    check = subprocess.run(["wget", "--spider", "-q", url])
    if check.returncode != 0 : #normaly 0
        return protein, False  # MSA not found


    subprocess.run(["wget", "-q", "-O", msa_in, url], check=True)

    nbr_line = sum(1 for _ in open(msa_in))
    if nbr_line < 3 : #if MSA have only 1 sequence, not useful for AF2 prediction, so generated it with mmseqs2
        return protein, False

    SP = len(prot_SP[protein]) - len(prot_no_SP[protein])
    if SP > 0 : #if no SP don't modify the MSA
        trimmed_records = []
        for rec in SeqIO.parse(msa_in, "fasta") :
            new_seq = rec.seq[SP:]
            len_aa_count = sum(1 for c in new_seq if c.isupper())
            if len_aa_count < 10 or not any(c.isupper() for c in str(new_seq)) : #skip short sequence after SP cut or sequence don't have aa
                continue
            new_rec = rec[:]
            new_rec.seq = new_seq
            trimmed_records.append(new_rec)

        if trimmed_records :
            with open(msa_in, "w") as msa_file :
                for rec in trimmed_records:
                    msa_file.write(f">{rec.description}\n{rec.seq}\n")

        #realign with mafft
        cmd_mafft = f"mafft --quiet --anysymbol --thread 1 --parttree --retree 1 --maxiterate 0 {msa_in} > {aln_out}" #monothread
        subprocess.run(cmd_mafft, shell=True, check=True)

        #reformat a3m
        cmd_reformat = f"reformat.pl fas a3m {aln_out} {msa_in}"
        subprocess.run(cmd_reformat, shell=True, check=True)
        cmd_rm_f = f"rm {aln_out}"
        subprocess.run(cmd_rm_f, shell=True, check=True)

        #delete \n
        all_lines = ""
        with open(msa_in, "r") as in_a3m:
            seq, header = "", None
            for line in in_a3m:
                if line.startswith(">"):
                    if header:
                        all_lines += f"{header}\n{seq}\n"
                    header = line.strip()
                    seq = ""
                else:
                    seq += line.strip()
            if header:
                all_lines += f"{header}\n{seq}\n"
        with open(msa_in, "w") as out_a3m:
            out_a3m.write(all_lines)

    return protein, True


def filter_signalP(file, Informations_dict, need_msa, need_pkl) :
    """
    Filter proteins based on the presence of a signal peptide using SignalP results.

    - Checks the first amino acid of each protein sequence to determine if a signal peptide is present (assumes sequences without SP start with 'M')
    - Updates the result dictionary with "Signal_peptide" status
    - Filters proteins from MSA and feature generation lists based on the user-specified SignalP
    - Keeps all proteins if no filtering criterion is specified

    Parameters :
    ----------
    file : object of class File_proteins
    Informations_dict : dict
    need_msa : list
    need_pkl : list

    Returns :
    ----------
    need_msa : list
    need_pkl : list
    """
    result_dict = file.get_result_dict()
    possible_prey = file.get_possible_prey()
    new_possible_prey = list()
    SignalP = Informations_dict["Signal_peptide"]
    for protein in possible_prey :
        if SignalP == "None" :
            new_possible_prey.append(protein)
        elif SignalP == result_dict[protein]["Signal_peptide"] :
            new_possible_prey.append(protein)
        elif SignalP != result_dict[protein]["Signal_peptide"] :
            result_dict[protein]["Reason_for_filtering"] = "Signal peptide : SignalP"
            if protein in need_msa :
                need_msa.remove(protein)
            if protein in need_pkl :
                need_pkl.remove(protein)
    if SignalP != "None" :
        logger.info("Protein preys remaining after SignalP filtering : " + str(len(new_possible_prey)))
    file.set_possible_prey(new_possible_prey)
    file.set_result_dict(result_dict)
    return need_msa, need_pkl

def Make_all_MSA_coverage(file, Path_Pickle_Feature, baits) :
    """
    Generate MSA coverage plots for all proteins and create a shallow_MSA summary file.

    - Computes sequence coverage and identity from the MSA
    - Generates a coverage heatmap saved as PDF
    - Determines if the MSA is "shallow" or missing (based on number of sequences)
    - Updates the result dictionary and possible prey list accordingly
    - Writes a summary of shallow MSAs to "log_file/shallow_MSA.txt"

    Parameters :
    ----------
    file : object of class File_proteins
    Path_Pickle_Feature : string
    baits : list
    """
    shallow_MSA = str()
    result_dict = file.get_result_dict()
    all_prot = file.get_possible_prey()
    if baits != [''] :
        for bait in baits :
            all_prot.append(bait)
    new_possible_prey = list()
    for prot in all_prot :
        if Path(f'{Path_Pickle_Feature}/{prot}_coverage.pdf').exists() == False :
            pre_feature_dict = pickle.load(open(f'{Path_Pickle_Feature}/{prot}.pkl','rb'))
            feature_dict = pre_feature_dict.feature_dict
            msa = feature_dict['msa']
            seqid = (np.array(msa[0] == msa).mean(-1))
            seqid_sort = seqid.argsort()
            non_gaps = (msa != 21).astype(float)
            non_gaps[non_gaps == 0] = np.nan
            final = non_gaps[seqid_sort] * seqid[seqid_sort, None]
            plt.figure(figsize=(14, 4), dpi=100)
            plt.subplot(1, 2, 1)
            plt.title(f"Sequence coverage ({prot})")
            plt.imshow(final, interpolation='nearest', aspect='auto', cmap="rainbow_r", vmin=0, vmax=1, origin='lower')
            plt.plot((msa != 21).sum(0), color='black')
            plt.xlim(-0.5, msa.shape[1] - 0.5)
            plt.ylim(-0.5, msa.shape[0] - 0.5)
            plt.colorbar(label="Sequence identity to query", )
            plt.xlabel("Positions")
            plt.ylabel("Sequences")
            plt.savefig(f"{Path_Pickle_Feature}/{prot+('_' if prot else '')}coverage.pdf")
            plt.close()
        with open(f'{Path_Pickle_Feature}/{prot}.a3m') as msa_f:
            line_msa = sum(1 for _ in msa_f)
        if line_msa/2 <= 100 and prot not in baits :
            if line_msa/2 < 2 :
               shallow_MSA += prot + " : " + str(int(line_msa/2)) + " sequences\n"
               result_dict[prot]["shallow_MSA"] = "No MSA"
               result_dict[prot]["Reason_for_filtering"] = "No MSA"
            else :
               shallow_MSA += prot + " : " + str(int(line_msa/2)) + " sequences\n"
               result_dict[prot]["shallow_MSA"] = "Shallow MSA"
               new_possible_prey.append(prot)
        if line_msa/2 > 100 and prot not in baits :
            new_possible_prey.append(prot)
        if line_msa/2 <= 100 and prot in baits :
            logger.warning(f"This bait : {prot} have a shallow MSA with {int(line_msa/2)} sequences. All interactions with this bait can be false negative.")
    with open("log_file/shallow_MSA.txt", "w") as MSA_file :
        MSA_file.write(shallow_MSA)
    file.set_possible_prey(new_possible_prey)
    file.set_result_dict(result_dict)


def filter_lenght(file, Informations_dict, need_msa, need_pkl, need_DeepLoc) :
    """
    Filter proteins based on their sequence length and update the list of possible preys.

    - Uses the full sequence (including signal peptide) to evaluate protein length
    - Filters proteins that are either too short or too long according to user-defined thresholds
    - Updates the result dictionary with a reason for filtering
    - Removes filtered proteins from MSA, pickle, and DeepLoc processing lists
    - Updates the File_proteins object with the new prey list

    Parameters :
    ----------
    file : object of class File_proteins
    Informations_dict : dict
    need_msa : list
    need_pkl : list
    need_DeepLoc : list

    Returns :
    ----------
    need_msa : list
    need_pkl : list
    need_DeepLoc : list
    """
    result_dict = file.get_result_dict()
    sequence_dict = file.get_proteins_sequence_SP() #use sequence with SP for lenght filtering
    possible_prey = file.get_possible_prey()
    if Informations_dict["Max_protein_lenght"] == "" : #not set by default, depend of GPU memory
        max_lenght = 100000
    else :
        max_lenght = int(Informations_dict["Max_protein_lenght"])
    min_lenght = int(Informations_dict["Min_protein_lenght"]) #set by default to 20
    new_possible_prey = list()
    for protein in possible_prey :
        if len(sequence_dict[protein]) < max_lenght and len(sequence_dict[protein]) > min_lenght :
            new_possible_prey.append(protein)
        else :
            result_dict[protein]["Reason_for_filtering"] = "Lenght filtering"
            if protein in need_msa :
                need_msa.remove(protein)
            if protein in need_pkl :
                need_pkl.remove(protein)
            if protein in need_DeepLoc :
                need_DeepLoc.remove(protein)
    logger.info("Protein preys remaining after lenght filtering : " + str(len(new_possible_prey)))
    file.set_possible_prey(new_possible_prey)
    file.set_result_dict(result_dict)
    return(need_msa, need_pkl, need_DeepLoc)


def filter_deeploc(file, Informations_dict, need_msa, need_pkl) :
    """
    Filter proteins based on predicted cellular localization and update the prey list as well as the MSA and pickle processing lists.

    - Compares each protein's predicted DeepLoc localization(s) with the user-specified allowed localizations
    - Keeps proteins matching at least one allowed localization
    - Marks proteins that do not match as filtered in the result dictionary
    - Updates the need_msa and need_pkl lists to include only proteins that pass the filter or are baits

    Parameters :
    ----------
    file : object of class File_proteins
    Informations_dict : dict
    need_msa : list
    need_pkl : list

    Returns :
    ----------
    new_need_msa : list
    new_need_pkl : list
    """
    localisation = Informations_dict["DeepLoc"].split(",")
    deeploc = file.get_deeploc()
    possible_baits = Informations_dict["Interact_with"]
    possible_prey = file.get_possible_prey()
    result_dict = file.get_result_dict()
    new_possible_prey = list()
    new_need_msa = list()
    new_need_pkl = list()
    for protein in possible_prey :
        for loc in deeploc[protein].split("|") :
            if loc in localisation :
                new_possible_prey.append(protein)
                break #stop the loop if one localisation is correct
        if protein not in new_possible_prey :
            result_dict[protein]["Reason_for_filtering"] = "Cellular localisation : DeepLoc"
    for prot_msa in need_msa :
        if prot_msa in new_possible_prey or prot_msa in possible_baits :
            new_need_msa.append(prot_msa)
    for prot_pkl in need_pkl :
        if prot_pkl in new_possible_prey or prot_pkl in possible_baits :
            new_need_pkl.append(prot_pkl)
    logger.info("Protein prey remaining after DeepLoc filtering : " + str(len(new_possible_prey)))
    file.set_possible_prey(new_possible_prey)
    file.set_result_dict(result_dict)
    return(new_need_msa, new_need_pkl)


def Generate_scripts(file, Informations_dict, Interaction_file, bait) :
    """
    Prepare the list of interaction jobs, taking into account protein lengths and GPU VRAM constraints to avoid out-of-memory (OOM) errors.

    - Estimates the maximum number of amino acids allowed based on available GPU VRAM
    - Builds interaction jobs (PPI or homo-oligomeric) only if they fit in memory
    - Skips interactions already computed
    - Records interactions that are too large for the GPU in a log file

    Parameters :
    ----------
    file : object of class File_proteins
    Informations_dict : dict
    Interaction_file : str
    bait : str

    Returns :
    ----------
    job_with_vram_length : list of tuple
    """
    job_with_vram_length = [] # [(job_str, int_length)]
    OOM_int = ""
    result_dict = file.get_result_dict()
    AF_version = Informations_dict["AlphaFold"]
    possible_prey = file.get_possible_prey()
    lenght_prot = file.get_lenght_prot()
    regions = Informations_dict["Regions"]

    # Estimate max amino acids based on GPU VRAM (choose first GPU as reference)
    pynvml.nvmlInit()
    handle = pynvml.nvmlDeviceGetHandleByIndex(0)
    mem_info = pynvml.nvmlDeviceGetMemoryInfo(handle)
    vram = (mem_info.total / 1024**2) * 0.001  # GiB
    max_aa=int((-0.00262+math.sqrt((0.00262**2)-4*0.00000228*(3-vram)))/(2*0.00000228))

    pynvml.nvmlShutdown()
    
    if Interaction_file == "PPI_int" :
        # Determine bait length and region
        complexe = "," in bait
        if complexe :
            save_multimer = bait
            bait_file = save_multimer.replace(",", "_and_")
            bait_for_job = save_multimer.replace(",", ";")
            lenght = sum(lenght_prot[prot] for prot in save_multimer.split(","))
            for prot in save_multimer.split(",") :
                if regions[prot] != "0-0":
                    start, end = int(regions[prot].split("-")[0]), int(regions[prot].split("-")[1])
                    bait_file = bait_file.replace(prot, f"{prot}_{start}-{end}")
                    bait_for_job = bait_for_job.replace(prot, f"{prot},{start}-{end}")
        else :
            lenght = lenght_prot[bait]
            if regions[bait] != "0-0" :
                start, end = int(regions[bait].split("-")[0]), int(regions[bait].split("-")[1])
                bait_file = f"{bait}_{start}-{end}"
                bait_for_job = f"{bait},{start}-{end}"
            else: 
                bait_file = bait
                bait_for_job = bait

        # Build job list
        copy_possible_prey = copy.deepcopy(possible_prey)  # To avoid modifying the list while iterating
        for prey in copy_possible_prey :
            int_lenght = lenght + lenght_prot[prey]

            # Check if model already exists
            if AF_version == "3":
                path1 = glob.glob(f"./result_PPI_int/{bait_file}_and_{prey}/ranked_0_model.cif")
                path2 = glob.glob(f"./result_PPI_int/{prey}_and_{bait_file}/ranked_0_model.cif")
            else : # AF_version == "2"
                path1 = glob.glob(f"./result_PPI_int/{bait_file}_and_{prey}/ranked_0.pdb")
                path2 = glob.glob(f"./result_PPI_int/{prey}_and_{bait_file}/ranked_0.pdb")

            if len(path1) == 0 and len(path2) == 0 :
                if int_lenght <= max_aa :
                    job_str = f"{bait_for_job};{prey}\n"
                    vram_lenght = 3 + 0.00262 * int_lenght + 0.00000228 * int_lenght**2
                    job_with_vram_length.append((job_str, vram_lenght))
                else :
                    OOM_int += f"{bait_for_job};{prey}\n"
                    result_dict[prey][f"Reason_for_filtering"] = "Interaction too large for your GPU, possible prey"
                    possible_prey.remove(prey)

    elif Interaction_file == "homo_int" :
        nbr_oligo = Informations_dict.get("Homo-oligomer", 2)
        for prey in possible_prey :
            int_lenght = lenght_prot[prey] * int(nbr_oligo)
            path = glob.glob(f"./result_homo_int/{prey}_homo_{nbr_oligo}er/ranked_0*")
            if len(path) == 0 :
                if int_lenght <= max_aa :
                    job_str = f"{prey}:{nbr_oligo}\n"
                    vram_lenght = 3 + 0.00262 * int_lenght + 0.00000228 * int_lenght**2
                    job_with_vram_length.append((job_str, vram_lenght))
                else :
                    OOM_int += f"{prey}:{nbr_oligo}\n"
                    result_dict[prey]["Reason_for_filtering"] = "Homo-oligomer too large for your GPU"
                    possible_prey.remove(prey)

    # Classify job_list in function of int lenght
    job_with_vram_length.sort(key=lambda x: x[1], reverse=True)


    # Save OOM interactions
    with open("log_file/OOM_interactions.txt", "w") as OOM_file :
        OOM_file.write(OOM_int)

    file.set_possible_prey(possible_prey)
    file.set_result_dict(result_dict)

    return job_with_vram_length





def Generate_3D_model(Informations_dict, interaction_type, job_with_vram_length, GPU, multi_job_per_gpu) :
    """
    Generate 3D models using multiple GPUs and multiprocessing.

    Parameters :
    ----------
    Informations_dict : dict
    interaction_files : str
    job_with_vram_length : list of tuple
    GPU : list
    multi_job_per_gpu : str
    """
    if not job_with_vram_length :
        return

    pynvml.nvmlInit()
    handle = pynvml.nvmlDeviceGetHandleByIndex(int(GPU[0]))
    mem_info = pynvml.nvmlDeviceGetMemoryInfo(handle)
    max_vram = mem_info.total / 1024**3 #Gb
    pynvml.nvmlShutdown()

    stop_flag = multiprocessing.Event()
    monitor = multiprocessing.Process(
        target=monitor_vram,
        args=(GPU, stop_flag)
    )
    monitor.start()

    AF_version = Informations_dict["AlphaFold"]
    Path_AlphaFold_Data = Informations_dict["Path_AlphaFold_Data"]
    Path_Pickle_Feature = Informations_dict["Path_Pickle_Feature"]

    with open("log_file/All_PPI_jobs.txt", "w") as f : #write complete script for informations
        for job, _ in job_with_vram_length :
            f.write(job + "\n")

    manager(jobs_pending=list(job_with_vram_length),
            GPU=GPU,
            max_vram=max_vram,
            Path_AlphaFold_Data=Path_AlphaFold_Data,
            Path_Pickle_Feature=Path_Pickle_Feature,
            interaction_type=interaction_type,
            AF_version=AF_version,
            multi_job_per_gpu=multi_job_per_gpu
        )



def manager(jobs_pending, GPU, max_vram, Path_AlphaFold_Data, Path_Pickle_Feature, interaction_type, AF_version, multi_job_per_gpu) :
    """
    Manage repartition of jobs on GPUs, monitor VRAM usage, and handle job completion.

    Parameters :
    ----------
    jobs_pending : list
    GPU : list
    max_vram : float
    Path_AlphaFold_Data : str
    Path_Pickle_Feature : str
    interaction_type : str
    AF_version : str
    multi_job_per_gpu : str
    """
    manager_proc = multiprocessing.Manager()
    result_queue = manager_proc.Queue()
    multi_job_per_gpu = False if multi_job_per_gpu == "False" else True
    gpu_vram_used = {gpu: 0.0 for gpu in GPU}
    jobs_running = {gpu: [] for gpu in GPU}  # {gpu_id: {job_str: process}}

    try :
        while jobs_pending or any(jobs_running.values()):

            while not result_queue.empty() :
                status, job, gpu_id, vram = result_queue.get()[:4]
                gpu_vram_used[gpu_id] -= vram
                for i, (j, p) in enumerate(jobs_running[gpu_id]):
                    if j == job:
                        p.join() 
                        jobs_running[gpu_id].pop(i)
                        break


            launched = False

            for interaction, job_vram in list(jobs_pending) :
                sorted_gpus = sorted(GPU, key=lambda g: gpu_vram_used[g]) #make PPI on GPU with minimum vram use

                for gpu_id in sorted_gpus :

                    free = max_vram - gpu_vram_used[gpu_id]
                    
                    if job_vram <= free and (multi_job_per_gpu or len(jobs_running[gpu_id]) == 0) :
                        p = multiprocessing.Process(target=gpu_job_runner,
                            args=(gpu_id,
                                interaction,
                                job_vram,
                                result_queue,
                                Path_AlphaFold_Data,
                                Path_Pickle_Feature,
                                interaction_type,
                                AF_version,
                            ),
                        )

                        p.start()

                        gpu_vram_used[gpu_id] += job_vram
                        jobs_running[gpu_id].append((interaction, p))
                        jobs_pending.remove((interaction, job_vram))
                        launched = True
                        break

                if launched :
                    break
            if not launched :
                time.sleep(1)

    except KeyboardInterrupt :
        kill_hint_processes()

def gpu_job_runner(gpu_id, interaction_file, vram, result_queue, Path_AlphaFold_Data, Path_Pickle_Feature, interaction_type, AF_version) :
    """
    Run a single AlphaFold job on a specified GPU, monitor its completion, and report results.

    Parameters :
    ----------
    gpu_id : int
    interaction_file : str
    vram : float
    result_queue : multiprocessing.Queue
    Path_AlphaFold_Data : str
    Path_Pickle_Feature : str
    interaction_type : str
    AF_version : str
    """
    env = os.environ.copy()
    env['CUDA_VISIBLE_DEVICES'] = str(gpu_id)
    env['XLA_PYTHON_CLIENT_PREALLOCATE'] = 'false'
    env['TF_FORCE_UNIFIED_MEMORY'] = 'true'
    env['XLA_PYTHON_CLIENT_MEM_FRACTION'] = '3.2'
    env['XLA_FLAGS'] = '--xla_gpu_enable_triton_gemm=false'

    prot_int = interaction_file.strip("\n").replace(";", "_and_")
    file_name = f"{interaction_type}_GPU_{prot_int}.txt"

    try :
        with open(f"log_file/{file_name}", "w") as f :
            f.write(interaction_file)

        if AF_version == "2" :
            cmd = (
                "run_multimer_jobs.py --mode=custom "
                "--num_cycle=3 "
                "--num_predictions_per_model=1 "
                "--compress_result_pickles=True "
                f"--output_path=./result_{interaction_type} "
                f"--data_dir={Path_AlphaFold_Data} "
                f"--protein_lists=log_file/{file_name} "
                f"--monomer_objects_dir={Path_Pickle_Feature} "
                "--remove_keys_from_pickles=False"
            )
        else :
            cmd = (
                "run_multimer_jobs.py --mode=custom "
                f"--output_path=./result_{interaction_type} "
                f"--data_dir={Path_AlphaFold_Data} "
                f"--protein_lists=log_file/{file_name} "
                f"--monomer_objects_dir={Path_Pickle_Feature} "
                "--fold_backend=alphafold3"
            )
        logger.info(f"[GPU {gpu_id}] Starting {interaction_file}")
        proc = subprocess.Popen(
            cmd,
            shell=True,
            env=env,
            preexec_fn=os.setsid
        )

        while True :

            ret = proc.poll()
            if ret is not None:
                if ret != 0 :
                    raise RuntimeError(f"Crash ({ret})")
                break

            time.sleep(0.5)

        result_queue.put(("done", interaction_file, gpu_id, vram))
        os.system(f"rm log_file/{file_name}")
        logger.info(f"[GPU {gpu_id}] Finished {interaction_file}")

    except Exception as e :
        result_queue.put(("Crash", interaction_file, gpu_id, vram, str(e)))



def monitor_vram(GPU, stop_flag) :
    """
    Monitor VRAM usage for given GPUs every `interval` seconds.
    """
    
    pynvml.nvmlInit()
    handles = {int(gpu): pynvml.nvmlDeviceGetHandleByIndex(int(gpu)) for gpu in GPU}

    try:
        while not stop_flag.is_set():
            lines = []
            for gpu, handle in handles.items():
                mem = pynvml.nvmlDeviceGetMemoryInfo(handle)
                used = mem.used / 1024**3
                total = mem.total / 1024**3
                pct = (used / total) * 100
                lines.append(f"GPU {gpu}: {used:.1f}/{total:.1f} GiB ({pct:.0f}%)")

            logger.info(" | ".join(lines))
            time.sleep(60)

    finally:
        pynvml.nvmlShutdown()


@staticmethod
def kill_hint_processes() :
    """
    Kill all processes related to the current Python executable, typically used to terminate any remaining AlphaFold jobs after a KeyboardInterrupt (Ctrl+C).
    """
    logger.info("Ctrl+C detected")
    python_path = sys.executable
    try :
        output = subprocess.check_output(f"pgrep -f '{python_path}'", shell=True).decode().split() #list all PID
    except subprocess.CalledProcessError :
        output = []

    for pid_str in output :
        pid = int(pid_str)
        try :
            os.killpg(os.getpgid(pid), signal.SIGKILL) #kill all PID
        except ProcessLookupError :
            pass



#Adapted from clean_pdb.py (https://gitlab.com/topf-lab/pi_score/-/tree/master/score_scripts)

from Bio.PDB import *
import sys
import os

dict_aa = {'CYS': 'C', 'ASP': 'D', 'SER': 'S', 'GLN': 'Q', 'LYS': 'K',
           'ILE': 'I', 'PRO': 'P', 'THR': 'T', 'PHE': 'F', 'ASN': 'N',
           'GLY': 'G', 'HIS': 'H', 'LEU': 'L', 'ARG': 'R', 'TRP': 'W',
           'ALA': 'A', 'VAL':'V', 'GLU': 'E', 'TYR': 'Y', 'MET': 'M'}

def str_object(pdbfile):
    '''
    Given a pdb file creates a structure object
    '''
    parser = PDBParser(QUIET=True)
    s = parser.get_structure('id', pdbfile)
    return s

class SelectResidues(Select):
    """ Only accept the specified residues when saving. """
    def accept_residue(self, residue):
        #print residue.get_resname(),residue.get_full_id()[3][0]
        if residue.get_full_id()[3][0] != 'W' and not residue.get_full_id()[3][0].startswith('H_'):
            if residue.get_resname() in dict_aa.keys():
                return 1

    def accept_atom(self,atom):
        if not atom.is_disordered() or atom.get_altloc() == 'A':
                if atom.element.strip() != 'H':
                        return 1

def change_atloc(pdbfile,
                out=None):
    if out == None:
        out = (pdbfile.rsplit('.')[0]
                + '_atm_loc'
                + '.pdb')
    
    outfle = open(out,'w')
    with open(pdbfile) as pdb:
        for lne in pdb:
            try:
                if len(lne.split()[3]) >3:
                    lne = lne.replace(lne.split()[3],' ' + lne.split()[3][1:])
                elif len(lne.split()[2]) > 3:
                    for keys in dict_aa.keys():
                        if keys in lne.split()[2]:
                            lne = lne.replace(lne.split()[2], lne.split()[2][0:3] + ' ' + lne.split()[2][4:])
            except: pass
            outfle.write(lne)
    outfle.close()
    return out

def write_sel_pdb(pdbfile):
    writer = PDBIO()
    struct = str_object(pdbfile)
    writer.set_structure(struct)
    outfle = (pdbfile.rsplit('.',1)[0]
                + '_clean'
                + '.pdb')
    writer.save(outfle, select=SelectResidues())

    return outfle


#Adapted from interface_assess.py (https://gitlab.com/topf-lab/pi_score/-/tree/master/score_scripts)

from Bio.PDB.MMCIFParser import MMCIFParser
import sys
from Bio.PDB import Selection, NeighborSearch, PDBParser, Select
import json

dict_aa = {'CYS': 'C', 'ASP': 'D', 'SER': 'S', 'GLN': 'Q', 'LYS': 'K',
           'ILE': 'I', 'PRO': 'P', 'THR': 'T', 'PHE': 'F', 'ASN': 'N', 
           'GLY': 'G', 'HIS': 'H', 'LEU': 'L', 'ARG': 'R', 'TRP': 'W', 
           'ALA': 'A', 'VAL':'V', 'GLU': 'E', 'TYR': 'Y', 'MET': 'M'}

class SelectChains(Select):
    """ Only accept the specified chains when saving. """
    def __init__(self, chain_letters):
        self.chain_letters = chain_letters

    def accept_chain(self, chain):
        return (chain.get_id() in self.chain_letters)

def str_object(pdbfile):
    '''
    Given a pdb file creates a structure object
    '''
    if pdbfile.rsplit('.')[-1] == 'pdb':
        parser = PDBParser(QUIET=True)
    elif pdbfile.rsplit('.')[-1] == 'ent':
        parser = PDBParser(QUIET=True)
    elif pdbfile.rsplit('.')[0] == 'cif':
        parser = MMCIFParser()
    else:
        print 'Not a valid PDB format'
        sys.exit()
    structure = parser.get_structure('id', pdbfile)
    return structure

def get_num_chains(structure):
    '''
    Returns a list of chain objects in a structure
    '''
    ch_list = Selection.unfold_entities(structure, 'C')
    return ch_list

def get_CAatoms(chain_object):
    '''
    for a given chain object, returns a list of CA atoms
    '''
    ca_list = []
    atm_list = Selection.unfold_entities(chain_object, 'A')
    for a in atm_list:
        if a.get_id() == 'CA':
            ca_list.append(a)
    return ca_list
    
def interface_residues(pdbfile,
                    len_chains,
                    numres_cut,
                    dist_cutoff):
    '''
    Given a PDB file calculates interface residues
    '''
#     numres_cut = int(numres_cut)
    structure = str_object(pdbfile)
    ch_list = get_num_chains(structure)
    int_reslist_j = []
    int_reslist_i = []
    dict_intf = {}
    dict_residue_contacts = {}
    for i in xrange(0,len(ch_list)-1):
        for j in xrange(i+1,len(ch_list)):
            int_reslist_j = []
            int_reslist_i = []
            intf_ch = ch_list[j].get_id() + '_' + ch_list[i].get_id()
            ca_i = get_CAatoms(ch_list[i])
            ca_j = get_CAatoms(ch_list[j])
            # print ca_i, intf_ch
            if len(ca_i) <len_chains or len(ca_j) <len_chains: 
                # dict_intf = {} 
                # dict_residue_contacts = {}
                continue
            ns_i = NeighborSearch(ca_i)
            for atms in ca_j:
                n = ns_i.search(atms.get_coord(), dist_cutoff,'R')
                if len(n) >0 :
                    res = (str(atms.get_parent().get_resname()) 
                             + str(atms.get_parent().get_id()[1]))
                    int_reslist_j.append(res)
                    t = []
                    for at in n:
                        res1 = (str(at.get_resname()) 
                             + str(at.get_id()[1]))
                        int_reslist_i.append(res1)
                        t.append(res1)
                    if intf_ch in dict_residue_contacts.keys():  
                        dict_residue_contacts[intf_ch].update({res:t})
                    else:
                        dict_residue_contacts[intf_ch] = {res:t}
            #Number of interface residues from each chain is >=10
            if len(set(int_reslist_j)) >= numres_cut and len(set(int_reslist_i)) >=numres_cut:
                dict_intf[intf_ch] = [list(set(int_reslist_j)),list(set(int_reslist_i))]   
    intf_outfle_prefix = pdbfile.split('.')[0]
    intf_outfle_name = intf_outfle_prefix + '_interface_residues_dict.json'
    
    if dict_intf: 
        with open(intf_outfle_name,'w') as outfle:
            json.dump(dict_intf,outfle)
    return dict_intf, dict_residue_contacts
    
    
def polar_residues(reslist):
    assert len(reslist) >0 
    polar_ref = ['SER', 
                 'THR',
                 'ASN',
                 'GLN',
                 'HIS',
                 'TYR']
    polar_reslist = []
    for r in reslist:
        if r in polar_ref:
            polar_reslist.append(r)
    frac_polar_residue = len(polar_reslist)/float(len(reslist))
    return frac_polar_residue

def hydrophobic_residues(reslist):
    assert len(reslist) >0 
    hydro_ref = ['ALA',
                 'LEU',
                 'ILE',
                 'VAL',
                 'PHE',
                 'TRP',
                 'CYS',
                 'MET'
                ]
    hydro_reslist = []
    for r in reslist:
        if r in hydro_ref:
            hydro_reslist.append(r)
    frac_hydro_residue = len(hydro_reslist)/float(len(reslist))
    return frac_hydro_residue
            
def charged_residues(reslist):
    assert len(reslist) >0 
    charged_ref = ['LYS',
                   'ARG',
                   'ASP',
                   'GLU']
    charged_reslist = []
    for r in reslist:
        if r in charged_ref:
            charged_reslist.append(r)
    frac_charged_residue = len(charged_reslist)/float(len(reslist))
    return frac_charged_residue


#!/usr/bin/env python3
#Adapted from get_good_inter_pae.py (https://github.com/KosinskiLab/AlphaPulldown/blob/main/alphapulldown/analysis_pipeline/alpha_analysis_jax0.4.def)

import sys
import os
sys.path.append(os.path.join(os.path.dirname(__file__), "script_pi_score"))
from calculate_mpdockq import *
import pickle
import json
import numpy as np
import pandas as pd
import subprocess
import gzip
from Bio.PDB import MMCIFParser, PDBIO
import gzip


def examine_inter_pae(pae_mtx,seqs,cutoff) :
    """A function that checks inter-pae values in multimer prediction jobs"""
    lens = [len(seq) for seq in seqs]
    old_lenth=0
    for length in lens:
        new_length = old_lenth + length
        pae_mtx[old_lenth:new_length,old_lenth:new_length] = 50
        old_lenth = new_length
    check = np.where(pae_mtx<cutoff)[0].size !=0

    return check


def obtain_mpdockq(work_dir,pkl_dict=None) :
    """Returns chain_coords,chain_CB_inds,plddt_per_chain,best_plddt,pdb_path"""
    pdb_path = os.path.join(work_dir,'ranked_0.pdb')
    pdb_chains, chain_coords, chain_CA_inds, chain_CB_inds = read_pdb(pdb_path)
    if pkl_dict ==  None :
        best_plddt = extract_plddt_from_pdb(pdb_path)
    else :
        best_plddt = pkl_dict['plddt'] 
    plddt_per_chain = read_plddt(best_plddt,chain_CA_inds)
    return chain_coords,chain_CB_inds,plddt_per_chain,best_plddt,pdb_path

def obtain_mpdockq2(chain_coords,chain_CB_inds,plddt_per_chain,best_plddt,pdb_path) :
    """Returns mpDockQ if more than two chains otherwise return pDockQ"""
    complex_score,num_chains = score_complex(chain_coords,chain_CB_inds,plddt_per_chain)
    if complex_score is not None and num_chains>2:
        mpDockq_or_pdockq = calculate_mpDockQ(complex_score)
    elif complex_score is not None and num_chains==2:
        chain_coords,plddt_per_chain = read_pdb_pdockq(pdb_path)
        mpDockq_or_pdockq = calc_pdockq(chain_coords,plddt_per_chain,t=8)
    else:
        mpDockq_or_pdockq = "None"
    return mpDockq_or_pdockq


def extract_plddt_from_pdb(pdb_file):
    """
    Extract plddt from b-factor in pdb
    """
    plddt_values = []
    with open(pdb_file, "r") as f:
        for line in f:
            if line.startswith("ATOM") and line[12:16].strip() == "CA":
                try:
                    b_factor = float(line[60:66].strip())
                    plddt_values.append(b_factor)
                except ValueError:
                    pass
    return np.array(plddt_values, dtype=float)

def run_and_summarise_pi_score(jobs, surface_thres, ccp4_setup) :
    """
    A function to calculate all predicted models' pi_scores and make a pandas df of the results.
    Instrumented to log timing per major step.
    """
    direc = jobs[0].split("/")[len(jobs[0].split("/"))-1]
    if os.path.isdir("/scratch") :
        tmp_dir = f"/scratch/tmp/{direc}"
    else :
        tmp_dir = f"/tmp/{direc}"
    print (f"Creating temporary directory {tmp_dir} for pi_score outputs")
    subprocess.run(f"rm -rf {tmp_dir} && mkdir -p {tmp_dir}/pi_score_outputs", shell=True, executable="/bin/bash", check=True)
    pi_score_outputs = os.path.join(tmp_dir, "pi_score_outputs")

    for job in jobs:
        if not os.path.isfile(os.path.join(job, "ranked_0.pdb")):
            print(f"{job} failed. Cannot find ranked_0.pdb in {job}")
            sys.exit()

        pdb_path = os.path.join(job, "ranked_0.pdb")
        output_dir = os.path.join(pi_score_outputs)
        cmd = (
            f"source {ccp4_setup}/bin/ccp4.setup-sh && "
            f"conda run -n pi_score python ./script_pi_score/run_piscore_wc.py "
            f"-p {pdb_path} -o {output_dir} -s {surface_thres} -ps 10"
        )

        proc = subprocess.Popen(cmd, shell=True, executable="/bin/bash", close_fds=True)
        proc.wait()
    output_df = pd.DataFrame()
    for job in jobs:
        name_job = job.split("/")[-1]
        subdir = os.path.join(pi_score_outputs)
        csv_files = [f for f in os.listdir(subdir) if 'filter_intf_features' in f]
        pi_score_files = [f for f in os.listdir(subdir) if 'pi_score_' in f]

        if not csv_files or not pi_score_files:
            print(f"Warning: missing CSV or pi_score files for {name_job}")
            continue
        for csv_f in csv_files:
            filtered_df = pd.read_csv(os.path.join(subdir, csv_f))

            if filtered_df.shape[0] == 0:
                for column in filtered_df.columns:
                    filtered_df[column] = ["None"]
                filtered_df['jobs'] = str(name_job)
                filtered_df['pi_score'] = "No interface detected"
                output_df = pd.concat([output_df, filtered_df])
                continue

            interface_id = csv_f.split("filter_intf_features_")[-1].replace(".csv", "")

            pi_f = [f for f in pi_score_files if interface_id in f]
            if not pi_f:
                print(f"Warning: no pi_score file for interface {interface_id}")
                continue

            pi_score = pd.read_csv(os.path.join(subdir, pi_f[0]))
            pi_score['jobs'] = str(name_job)

            # interface propre
            if 'chains' in pi_score.columns:
                pi_score['interface'] = pi_score['chains']
            else:
                pi_score['interface'] = interface_id

            filtered_df['jobs'] = str(name_job)
            last_chain = get_last_chain_from_pdb(os.path.join(job, "ranked_0.pdb"))
            last_chain = get_last_chain_from_pdb(os.path.join(job, "ranked_0.pdb"))

            filtered_df = filtered_df[filtered_df['interface'].apply(lambda x: last_chain in x)]
            pi_score = pi_score.drop(columns=["#PDB", "pdb", "pvalue", "chains", "predicted_class"],errors="ignore")

            filtered_df = pd.merge(
                filtered_df,
                pi_score,
                on=['jobs', 'interface'],
                how='left'
            )

            output_df = pd.concat([output_df, filtered_df])

    subprocess.run(f"rm -rf {tmp_dir}", shell=True, executable='/bin/bash')
    return output_df
    
    

def main(job, output_dir, cutoff, surface_thres, ccp4_setup, seq_no_SP ,AF_version) :
    good_jobs = []
    iptm_ptm = list()
    iptm = list()
    mpDockq_scores = list()
    logging.info(f"Scoring {job}")
    result_subdir = os.path.join(job)
    interaction = job.split("/")[-1]
    seqs = list()
    for prot in interaction.split("_and_") :
        if "-" in prot and prot.split("_")[0] in seq_no_SP.keys() :
           prot = prot.split("_")[0]
        seqs.append(seq_no_SP[prot])
    if AF_version == "3" : #for alphafold3
        if os.path.isfile(os.path.join(result_subdir,'ranked_0.pdb')) == False : #create ranked_0.pdb for AF3
            parser = MMCIFParser(QUIET=True)
            structure = parser.get_structure('model', os.path.join(result_subdir,'ranked_0_model.cif'))
            io = PDBIO()
            io.set_structure(structure)
            io.save(os.path.join(result_subdir,'ranked_0.pdb'))
        if os.path.isfile(os.path.join(job,'ranked_0_summary_confidences.json')) :
            with open(os.path.join(result_subdir,'ranked_0_summary_confidences.json'),'rb') as json_sum_f :
                json_sum = json.load(json_sum_f)
            if "iptm" in json_sum.keys() and "ptm" in json_sum.keys() :
                iptm_score = json_sum['iptm']
                ptm_score = json_sum['ptm']
                iptm_ptm_score = 0.8 * iptm_score + 0.2 * ptm_score
            with open(os.path.join(result_subdir, 'ranked_0_confidences.json'),'rb') as json_f :
                json_data = json.load(json_f)
            pae_list = json_data['pae']
            pae_mtx = np.array(pae_list)
            chain_coords,chain_CB_inds,plddt_per_chain,best_plddt,pdb_path = obtain_mpdockq(os.path.join(job))
            check = examine_inter_pae(pae_mtx,seqs,cutoff=cutoff)
            mpDockq_score = obtain_mpdockq2(chain_coords,chain_CB_inds,plddt_per_chain,best_plddt,pdb_path)
            if check:
                good_jobs.append(str(job))
                iptm_ptm.append(iptm_ptm_score)
                iptm.append(iptm_score)
                mpDockq_scores.append(mpDockq_score)
                


    if os.path.isfile(os.path.join(job,'ranking_debug.json')): #for alphafold2
        with open(os.path.join(result_subdir,'ranking_debug.json'),'rb') as json_f :
            data = json.load(json_f)
        best_model = data['order'][0]
        if "iptm" in data.keys() or "iptm+ptm" in data.keys():
            iptm_ptm_score = data['iptm+ptm'][best_model]
            if os.path.exists(os.path.join(result_subdir, f"result_{best_model}.pkl")) :
                pkl_path = os.path.join(result_subdir, f"result_{best_model}.pkl")
                with open(pkl_path, 'rb') as pkl :
                    check_dict = pickle.load(pkl)
            elif os.path.exists(os.path.join(result_subdir, f"result_{best_model}.pkl.gz")) :
                print("result pickle for the best model not found. Now search for zipped pickle.")
                pkl_path = os.path.join(result_subdir, f"result_{best_model}.pkl.gz")
                with gzip.open(pkl_path, 'rb') as pkl :
                    check_dict = pickle.load(pkl)
            else :
                logging.info(f"Cannot find result pickle for {job}, skipping.")
            iptm_score = check_dict['iptm']
            pae_mtx = check_dict['predicted_aligned_error']
            chain_coords,chain_CB_inds,plddt_per_chain,best_plddt,pdb_path = obtain_mpdockq(os.path.join(job),check_dict)
            check = examine_inter_pae(pae_mtx,seqs,cutoff=cutoff)
            mpDockq_score = obtain_mpdockq2(chain_coords,chain_CB_inds,plddt_per_chain,best_plddt,pdb_path)
            if check:
                good_jobs.append(str(job))
                iptm_ptm.append(iptm_ptm_score)
                iptm.append(iptm_score)
                mpDockq_scores.append(mpDockq_score)
    other_measurements_df=pd.DataFrame.from_dict({
        "jobs":job.split("/")[-1],
        "iptm_ptm":iptm_ptm,
        "iptm":iptm,
        "mpDockQ/pDockQ":mpDockq_scores})

    if good_jobs!=[] :
        pi_score_df = run_and_summarise_pi_score(good_jobs,surface_thres,ccp4_setup)
        pi_score_df = pd.merge(pi_score_df,other_measurements_df,on="jobs")
        columns = list(pi_score_df.columns.values)
        columns.pop(columns.index('jobs'))
        pi_score_df = pi_score_df[['jobs'] + columns]
        pi_score_df = pi_score_df.sort_values(by='iptm',ascending=False)
        return pi_score_df
    else :
        return None



def get_last_chain_from_pdb(pdb_file):
    last_chain = None
    with open(pdb_file, 'r') as f:
        for line in f:
            if line.startswith(('ATOM', 'HETATM')):
                chain = line[21]  # col 22 (index 21)
                last_chain = chain
    return last_chain



"""
Main entry point of HInt

Author: Quentin Rouger

This script orchestrates the full HInt pipeline, from input parsing to final results generation.
"""

import sys
import logging
import argparse
import os

# ------------------------------------------------------------------
# Logging configuration
# ------------------------------------------------------------------

log_filename = "./log_file/HInt.log"

# Create log directory if it does not exist
if not os.path.exists("log_file") :
    os.system("mkdir log_file")

from .Utils_HInt import *
from .File_proteins import *
from .Scoring_HInt import Score_interaction, Resume_file, Create_figures

# Reset existing handlers to avoid duplicated logs
logging.getLogger().handlers.clear()
logger = logging.getLogger()
logger.setLevel(logging.INFO)
formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
file_handler = logging.FileHandler(log_filename, mode='w')
file_handler.setFormatter(formatter)
stream_handler = logging.StreamHandler(sys.stdout)
stream_handler.setFormatter(formatter)
logger.addHandler(file_handler)
logger.addHandler(stream_handler)


# ------------------------------------------------------------------
# Argument parsing
# ------------------------------------------------------------------

def add_arguments(parser) :
    """
    Define command-line arguments.

    Parameters
    ----------
    parser : argparse.ArgumentParser
    """
    parser.add_argument("--gpu", help="Comma-separated list of GPUs available for computation", required=False, default="0")

    N_CPU = multiprocessing.cpu_count()
    default_cpu = N_CPU // 2  # by default, use half of the available CPUs
    parser.add_argument("--cpu", help="Number of CPUs available for computation", required=False, default=default_cpu, type=int)
    parser.add_argument("--multi_job_per_gpu", help="Allow multiple jobs to run on the same GPU if VRAM allows it (default: True)", required=False, default="True")


# ------------------------------------------------------------------
# Main execution
# ------------------------------------------------------------------

def main() :

    parser = argparse.ArgumentParser()
    add_arguments(parser)
    args = parser.parse_args()

    GPU = [gpu for gpu in args.gpu.split(",")]
    CPU = args.cpu
    multi_job_per_gpu = args.multi_job_per_gpu
    if multi_job_per_gpu not in ["True", "False"] :
        raise ValueError("Invalid value for --multi_job_per_gpu.")
        
    Informations_dict = Define_informations()

    logger.info("GPUs set to: %s", GPU)
    logger.info("Number of CPUs set to: %s", CPU)

    # --------------------------------------------------------------
    # Initialize protein container
    # --------------------------------------------------------------

    HInt_object = File_proteins(Informations_dict["Path_Uniprot_ID"])

    
    for bait in Informations_dict["Interact_with"] : # Check that all bait proteins exist in the protein list
        if bait not in HInt_object.get_proteins() and bait != "" :
            raise Exception(f"Bait {bait} not found in the protein list {Informations_dict['Path_Uniprot_ID']}")

    # --------------------------------------------------------------
    # Checkpointing: determine which features need to be computed
    # --------------------------------------------------------------


    # need_msa also includes proteins that only require signal peptide information
    need_msa, need_pkl, need_DeepLoc = HInt_object.check_save_dict(Informations_dict["Path_Pickle_Feature"])


    # Remove bait proteins from the prey list
    HInt_object.set_possible_prey([protein for protein in HInt_object.get_possible_prey() if protein not in Informations_dict["Interact_with"]])

    # --------------------------------------------------------------
    # Length-based filtering
    # --------------------------------------------------------------

    # Filter proteins based on sequence length
    # (default: remove proteins shorter than 20 AA)
    need_msa, need_pkl, need_DeepLoc = filter_lenght(HInt_object, Informations_dict, need_msa, need_pkl, need_DeepLoc)

    # --------------------------------------------------------------
    # DeepLoc filtering
    # --------------------------------------------------------------

    if len(need_DeepLoc) > 0 : # Run DeepLoc only for proteins without localization information
        run_deeploc(HInt_object, Informations_dict["Organism"], need_DeepLoc, GPU)


    if Informations_dict["DeepLoc"].split(",") != ["None"] : # Apply DeepLoc-based filtering if enabled
        need_msa, need_pkl = filter_deeploc(HInt_object, Informations_dict, need_msa, need_pkl)

    HInt_object.Make_save_dict()  # Save DeepLoc results

    # --------------------------------------------------------------
    # Signal peptide processing
    # --------------------------------------------------------------

    need_SP = list()
    for protein in need_msa :
        if protein not in HInt_object.get_proteins_sequence_no_SP().keys() or protein not in HInt_object.get_prot_SP().keys() : #protein need MSA but can already have sequence without SP
            need_SP.append(protein)
    if len(need_SP) > 0 : # Run SignalP for proteins without signal peptide annotation
        need_msa = run_SP(HInt_object, Informations_dict, need_SP, need_msa)

    HInt_object.Make_save_dict() # Save sequences without signal peptides

    
    for bait in Informations_dict["Interact_with"] : # Adjust bait protein lengths if specific regions are defined
        if Informations_dict["Regions"][bait] != "0-0" :
            dict_lenght = HInt_object.get_lenght_prot()
            start = int(Informations_dict["Regions"][bait].split("-")[0])
            end = int(Informations_dict["Regions"][bait].split("-")[1])
            dict_lenght[bait] = end - start + 1
            HInt_object.set_lenght_prot(dict_lenght)


    # Filter proteins based on signal peptide criteria
    need_msa, need_pkl = filter_signalP(HInt_object, Informations_dict, need_msa, need_pkl)


    # --------------------------------------------------------------
    # Feature generation (MSA + pickle files)
    # --------------------------------------------------------------


    if len(need_msa) > 0 or len(need_pkl) > 0 :
        create_feature(HInt_object, Informations_dict, GPU, CPU, need_msa, need_pkl)

    HInt_object.Make_save_dict()  # Save SignalP and feature results


    # --------------------------------------------------------------
    # MSA depth analysis
    # --------------------------------------------------------------


    logger.info("Generating MSA depth figures")
    Make_all_MSA_coverage(HInt_object, Informations_dict["Path_Pickle_Feature"], Informations_dict["Interact_with"])


    # --------------------------------------------------------------
    # Protein–protein interaction modeling
    # --------------------------------------------------------------


    if Informations_dict["Interact_with"] != [''] :
        for bait in Informations_dict["Multimer_bait"] :
            job_with_vram_length = Generate_scripts(HInt_object, Informations_dict, "PPI_int", bait)
            Generate_3D_model(Informations_dict, "PPI_int", job_with_vram_length, GPU, multi_job_per_gpu)
            Score_interaction(HInt_object, Informations_dict, CPU, "PPI_int", bait)
            HInt_object.Make_save_dict()  # Save interaction scores


    # --------------------------------------------------------------
    # Homo-oligomer modeling
    # --------------------------------------------------------------


    if int(Informations_dict["Homo-oligomer"]) > 1 :
        job_with_vram_length = Generate_scripts(HInt_object, Informations_dict, "homo_int", "")
        Generate_3D_model(Informations_dict, "homo_int", job_with_vram_length, GPU, multi_job_per_gpu)
        Score_interaction(HInt_object, Informations_dict, CPU, "homo_int")


    # --------------------------------------------------------------
    # Final summary and figures
    # --------------------------------------------------------------


    sorted_protein = Resume_file(HInt_object, Informations_dict)
    if Informations_dict["Interact_with"] != [''] :
        Create_figures(HInt_object, Informations_dict, Informations_dict["AlphaFold"], sorted_protein, CPU)



#Adapted from run_piscore_wc.py (https://gitlab.com/topf-lab/pi_score/-/tree/master/score_scripts)
#code in python 2.7
import argparse
import os
import shutil
from interface_assess import *
from sc_utils import write_sc_script,run_sc, parse_sc_output
from clean_pdb import write_sel_pdb, change_atloc
from pisa_utils import run_pisa, parse_pisa_xml_outfle
from pi_score_utils import write_csv_with_features_wc, make_predictions, filter_csv
import time
import sys
import uuid 

author = 'Author: Sony Malhotra '


parser = argparse.ArgumentParser(description="Calculate PI-score for a given complex structure")
group = parser.add_mutually_exclusive_group(required=True)
group.add_argument("-p", "--pdb", help="PDB file to calculate the PI-score", default=None)
group.add_argument("-d", "--dirc", help="directory to assess the PDB files",default=None)
parser.add_argument("-ch", "--chains", help="Chains forming the interface which needs to be assessed",default=None)
parser.add_argument("-dc", "--distance_cutoff", help="Distance cutoff for interface definition",default=7, type =float)
parser.add_argument("-s", "--intf_size", help="Number of interface residues from each of the interacting subunits",default=10,type=int)
parser.add_argument("-ps", "--prot_size", help="Length of each of the chains",default=30,type=int)
parser.add_argument("-np", "--num_processors", help="Number of processors to use",default=40,type=int)
parser.add_argument("-o", "--out", help="Name of the output directory to write results",default=os.path.join(os.getcwd(),'pi_output'))
parser.add_argument("-c","--csv",help="Name of the CSV files where features will be stored",default=None)
parser.add_argument("-m","--model",help="SVM model to make new predictions",choices=['A','B','WC'], default='WC')
parser.add_argument("-r","--results",help="Name of the file where model predictions will be written",default=None)

def main(): 
    args = parser.parse_args()
    pdblist = []
    
    if args.dirc != None:
        #print ('Setting chains to all chains')
        args.chains = None

    #print ('*****Selecting model for training*****')
    dirname, filename = os.path.split(os.path.abspath(__file__))
    if args.model == 'A':
        saved_model = os.path.join(os.path.dirname(sys.argv[0]),'svm_model/finalized_model_A.sav')
        saved_sc = os.path.join(dirname,'svm_model/scaler_model_A.sav')
    elif args.model == 'B':
        saved_model = os.path.join(os.path.dirname(sys.argv[0]),'svm_model/finalized_model_B.sav')
        saved_sc = os.path.join(dirname,'svm_model/scaler_model_B.sav')
    if args.model == 'WC':
        saved_model = os.path.join(os.path.dirname(sys.argv[0]),'svm_model/finalized_model_wc.sav')
        saved_sc = os.path.join(dirname,'svm_model/scaler_model_wc.sav')
    working_path = os.getcwd()

    #print ('*****Setting output directory for results*****')
    if not os.path.isdir(args.out):
        os.mkdir(args.out)

    session_id = uuid.uuid4().hex[:8]

    if args.csv is None:
        current_time = time.strftime("%m.%d.%y_%H%M", time.localtime())
        obj = session_id
        args.csv = os.path.join(args.out, "intf_features_%s_%s.csv" % (current_time, session_id))
        
    if args.results is None:
        current_time = time.strftime("%m.%d.%y_%H%M", time.localtime())
        args.results = os.path.join(args.out,"pi_score_%s_%s_id.txt" % (current_time, session_id))
    
    if os.path.isfile(args.results):
        os.remove(args.results)
    
    out = open(args.results,'w')
    out.write('#PDB,chains,predicted_class,pi_score\n')
    out.close()
    
    filter_csvfle = os.path.join(args.out,"filter_intf_features_%s_%s.csv" % (current_time, session_id))

    #print ('*****Setting input for assessment*****')
    #get a list of pdbfiles to score interfaces for
    if args.pdb:
        pdblist = [args.pdb.rsplit('/',1)[-1]]
    elif args.dirc:
        for fl in os.listdir(args.dirc):
            if fl.endswith('pdb'):
                pdblist.append(fl)
    else:
        print ('Not a valid input')


    #list of directories where to run conservation
    #dir_infiles = []
    #all_conslst = []
    #all_consdir = []
    #ch_lst = []

    #print ('*****Iterating over each structure to assess*****')
    for pdb in pdblist:
        chains_in = []
        os.chdir(working_path)
        pd = pdb.rsplit('.',1)[0]
        #make an output directory
        if os.path.isdir(os.path.join(args.out,pd)):
            shutil.rmtree(os.path.join(args.out,pd))
        #print ('*****Making results subdirectory specific to each structure in output directory*****')
        #print (os.path.join(args.out,pd))
        os.mkdir(os.path.join(args.out,pd))
        #copy input files
        if args.dirc:
            copy_cmd = 'cp '+ os.path.join(args.dirc,pdb) + ' ' + os.path.join(args.out,pd)
        else:
            copy_cmd = 'cp '+ os.path.join(args.pdb) + ' ' + os.path.join(args.out,pd)
        #print ('*****Doing PDB,',pdb,' *****')
        os.system(copy_cmd)
        os.chdir(os.path.join(args.out,pd))
        #print ('*****Calculating interface*****')
        #Interface calculation
        dict_intf,dict_contact = interface_residues(pdb,
                                                    len_chains=args.prot_size,
                                                    numres_cut=args.intf_size,
                                                    dist_cutoff=args.distance_cutoff)
        intf_outfle_prefix = pd
        if not dict_intf: 
            intf_outfle_name = intf_outfle_prefix + '_no_int.txt'
            a = open(intf_outfle_name,'w')
            a.write('No interfaces found')
            a.close()
            continue
        
        #print ('*****Calculating contacts*****')
        #Writing the contact dictionary, which residue from which chains are in contact
        contact_name = intf_outfle_prefix + '_interface_chain_contacts.json'
        with open(contact_name,'w') as outfle:
            json.dump(dict_contact,outfle)
        if args.chains:
            for c in args.chain:
                chains_in.append(c)
        dict_intf_prop = {}
        
        #Write clean pdbfile for shape complementarity
        clean_pdbfile = write_sel_pdb(pdb)
        atm_file = change_atloc(clean_pdbfile)
        os.system('rm -f ' + clean_pdbfile)
        os.system('mv ' + atm_file + ' ' + clean_pdbfile)
        
        #print ('*****Iterating over interfaces in structure*****')
        #Iterating over all the interfaces
        index = 0
        for k in dict_intf:
            index+=1
            ch1 = k.split('_')[0]
            ch2 = k.split('_')[-1]
            flag = False
            if args.chains:
                if ch1 in args.chains and ch2 in args.chains:
                    flag = True
                    break
                else:
                    print ('No interfaces between the input chains,', args.chains)
                    continue
            #print ('*****Calculating interface features*****')
            if flag or not args.chains: 
                ch1_intf_res = dict_intf[k][0]
                ch2_intf_res = dict_intf[k][1]
                dict_intf_prop[k] = {'Intfresidues_ch1':[ch1,len(ch1_intf_res)]}
                dict_intf_prop[k] = {'Intfresidues_ch2':[ch2,len(ch2_intf_res)]}
                total_intf_res = ch1_intf_res + ch2_intf_res
                total_num_intf_res = len(total_intf_res)
                only_res_name_lst = []
                for re in total_intf_res:
                    only_res_name_lst.append(re[0:3])
                polar = polar_residues(only_res_name_lst)
                hydro = hydrophobic_residues(only_res_name_lst)
                charged = charged_residues(only_res_name_lst)
                if dict_intf_prop[k]:
                    dict_intf_prop[k].update({'Polar':polar, 'Charged': charged, 'Hydrophobhic':hydro, 'Num_intf_residues': total_num_intf_res})
                else:
                    dict_intf_prop[k] = {'Polar':polar, 'Charged': charged, 'Hydrophobhic':hydro, 'Num_intf_residues': total_num_intf_res}
                intf_outfle_name = intf_outfle_prefix + '_interface_properties_dict.json'
                with open(intf_outfle_name,'w') as outfle:
                    json.dump(dict_intf_prop,outfle)     

                #print ('*****Calculating shape complementarity*****')
                #calculate shape complementarity
                scriptfile = write_sc_script(chain_lst=k.split('_'),
                                            pdbfile=clean_pdbfile)
                sc_dir = args.out+"/ranked_0/"+scriptfile
                run_sc(sc_dir)
                #parse_sc_output('tmp_sc.out')
                
        #print ('*****Calculating pisa *****')
        #calculate features using pisa
        outfle_name = run_pisa(clean_pdbfile)
        pisa_out = parse_pisa_xml_outfle(outfle_name)

    os.chdir(working_path)

    #print ('*****Writing CSV file with features *****')
    # # Write the CSV file with interface features
    
    write_csv_with_features_wc(indir=args.out, outfle=args.csv)
    # 
    # 
    # #filter CSV and have only interfaces where all features were computed successfully
    filter_csv(args.csv,filter_csvfle)

    #print ('*****Calculating PI-score*****')
    # # #Make predictions
    make_predictions(saved_sc=saved_sc, 
                    csvfile=filter_csvfle, 
                    saved_model=saved_model,
                    outfle=args.results,
                    session_id=session_id)
if __name__ == '__main__':
    main()




#Adapted from utils.py (https://gitlab.com/topf-lab/pi_score/-/tree/master/score_scripts)

import os
import json
import pandas as pd
import numpy as np
import pickle
import sys

def write_csv_with_features_wc_2(indir,outfle):
    out_csv = outfle
    lst_features = ['Num_intf_residues', 'Polar', 'Hydrophobhic', 'Charged']
    outfle = open(out_csv,'w')
    #writing header in csv file
    outfle.write('pdb,interface,')
    for features in lst_features:
        outfle.write(features + ',')
    outfle.write('contact_pairs')
    outfle.write(',sc,hb,sb,int_solv_en,int_area,pvalue')
    outfle.write('\n')
    for fle in os.listdir(indir):
        if os.path.isdir(os.path.join(indir,fle)):
            intf_dict = ''
            cons = {}
            pisa_dict = {}
            dict_num_contacts = {}
            sc_dict = {}

            for j in os.listdir(os.path.join(indir,fle)):
                nw_path = os.path.join(os.path.join(indir,fle),j)
                if not os.path.isfile(nw_path):
                    continue
                if not nw_path.endswith('json'):
                    continue

                if nw_path.endswith('_interface_chain_contacts.json'):
                    contacts_dict = nw_path
                    dict_num_contacts = count_contacts_at_interface(contacts_dict)

                elif nw_path.endswith('interface_properties_dict.json'):
                    intf_dict = nw_path

                elif nw_path.endswith('sc_scores.json'):
                    sc_dict1 = nw_path
                    with open(sc_dict1,'rb') as handle:
                        sc_dict = json.load(handle)
                elif nw_path.endswith('dict_pisa.json'):
                    with open(nw_path) as handle:
                        pisa_dict = json.load(handle)
                if intf_dict:
                    with open(intf_dict) as handle:
                        dict_intf_prop = json.load(handle)
                        for intf in dict_intf_prop:
                            outfle.write(fle + ',' + intf + ',')
                            for feature in lst_features:
                                if feature == lst_features[0]:
                                    outfle.write(str(dict_intf_prop[intf][feature]) + ',')
                                else:
                                    try:
                                        outfle.write(str(round(dict_intf_prop[intf][feature],3)) + ',')
                                    except:
                                        outfle.write('NA' + ',')
                            intf1 = intf.strip().split('_')[-1] + '_' + intf.strip().split('_')[0]
                            if dict_num_contacts:
                                outfle.write(str(dict_num_contacts[intf]) + ',')
                            else:
                                outfle.write('NA,')
                            if sc_dict:
                                for pdbid_sc in sc_dict.keys():
                                    if intf in sc_dict[pdbid_sc].keys() and len(sc_dict[pdbid_sc][intf]) > 0:
                                        outfle.write(str(sc_dict[pdbid_sc][intf]) + ',')
                                    elif intf1 in sc_dict[pdbid_sc].keys() and len(sc_dict[pdbid_sc][intf1]) > 0:
                                        outfle.write(str(sc_dict[pdbid_sc][intf1]) + ',')
                                    else:
                                        outfle.write('NA,')
                            else:
                                outfle.write('NA,')

                            if pisa_dict:
                                k = ''
                                if intf in pisa_dict.keys():
                                    k = intf
                                elif intf1 in pisa_dict.keys():
                                    k = intf1
                                if k:
                                    try:
                                        outfle.write(str(pisa_dict[k]['hb']) + ',')
                                    except:
                                        outfle.write('NA,')
                                    try:
                                        outfle.write(str(pisa_dict[k]['sb']) + ',')
                                    except:
                                        outfle.write('NA,')
                                    try:
                                        outfle.write(str(pisa_dict[k]['int_solv_en']) + ',')
                                    except:
                                        outfle.write('NA,')
                                    try:
                                        outfle.write(str(pisa_dict[k]['int_area']) + ',')
                                    except:
                                        outfle.write('NA,')
                                    try:
                                        outfle.write(str(pisa_dict[k]['pvalue']))
                                    except:
                                        outfle.write('NA')
                                else:
                                    outfle.write('NA,NA,NA,NA,NA')
                            outfle.write('\n')
    outfle.close()

def write_csv_with_features_wc(indir,outfle):
    out_csv = outfle
    lst_json_fles = []
    lst_features = ['Num_intf_residues', 'Polar', 'Hydrophobhic', 'Charged']
    outfle = open(out_csv,'w')
    #writing header in csv file
    outfle.write('pdb,interface,')
    for features in lst_features:
        outfle.write(features + ',')
    outfle.write('contact_pairs')
    outfle.write(', sc, hb, sb, int_solv_en, int_area, pvalue')
    outfle.write('\n')
    for fle in os.listdir(indir):
        if os.path.isdir(os.path.join(indir,fle)):
            intf_dict = ''
            cons = {}
            pisa_dict = {}
            dict_num_contacts = {}
            complete_sc_dict = {}

            for j in os.listdir(os.path.join(indir,fle)):
                nw_path = os.path.join(os.path.join(indir,fle),j)
                if not os.path.isfile(nw_path):
                    continue
                if not nw_path.endswith('json'):
                    continue

                if nw_path.endswith('_interface_chain_contacts.json'):
                    contacts_dict = nw_path
                    dict_num_contacts = count_contacts_at_interface(contacts_dict)

                elif nw_path.endswith('interface_properties_dict.json'):
                    intf_dict = nw_path

                elif nw_path.endswith('sc_scores.json'):
                    with open(nw_path,'rb') as handle:
                        sc_dict = json.load(handle)
                        for pdbid, intf_dict_sc in sc_dict.items():
                            if pdbid not in complete_sc_dict:
                                complete_sc_dict[pdbid] = {}
                            complete_sc_dict[pdbid].update(intf_dict_sc)

                elif nw_path.endswith('dict_pisa.json'):
                    with open(nw_path) as handle:
                        pisa_dict = json.load(handle)

            if intf_dict:
                with open(intf_dict) as handle:
                    dict_intf_prop = json.load(handle)
                    for intf in dict_intf_prop:
                        outfle.write(fle + ',' + intf + ',')
                        for feature in lst_features:
                            if feature == lst_features[0]:
                                outfle.write(str(dict_intf_prop[intf][feature]) + ',')
                            else:
                                try:
                                    outfle.write(str(round(dict_intf_prop[intf][feature],3)) + ',')
                                except:
                                    outfle.write('NA' + ',')
                        #if cons:
                        #    try:
                        #        outfle.write(str(round(cons[intf],3)) + ',')
                        #    except:
                        #        outfle.write('NA'+ ',')
                        #else:
                        #   outfle.write('NA'+ ',')
                        intf1 = intf.strip().split('_')[-1] + '_' + intf.strip().split('_')[0]
                        if dict_num_contacts:
                            outfle.write(str(dict_num_contacts[intf]) + ',')
                        else:
                            outfle.write('NA,')
                        sc_val = 'NA'
                        for pdbid_sc, intf_sc_dict in complete_sc_dict.items():
                            if intf in intf_sc_dict:
                                sc_val = intf_sc_dict[intf]
                                break
                            elif intf1 in intf_sc_dict:
                                sc_val = intf_sc_dict[intf1]
                                break
                        outfle.write(sc_val+",")

                        if pisa_dict:
                            k = ''
                            if intf in pisa_dict.keys():
                                k = intf
                            elif intf1 in pisa_dict.keys():
                                k = intf1
                            if k:
                                try:
                                    outfle.write(str(pisa_dict[k]['hb']) + ',')
                                except:
                                    outfle.write('NA,')
                                try:
                                    outfle.write(str(pisa_dict[k]['sb']) + ',')
                                except:
                                    outfle.write('NA,')
                                try:
                                    outfle.write(str(pisa_dict[k]['int_solv_en']) + ',')
                                except:
                                    outfle.write('NA,')
                                try:
                                    outfle.write(str(pisa_dict[k]['int_area']) + ',')
                                except:
                                    outfle.write('NA,')
                                try:
                                    outfle.write(str(pisa_dict[k]['pvalue']))
                                except:
                                    outfle.write('NA')
                            else:
                                outfle.write('NA,NA,NA,NA,NA')
                        outfle.write('\n')


def count_contacts_at_interface(contacts_dict):
    dict_num_contacts = {}
    with open(contacts_dict,'rb') as handle:
        dict_contacts = json.load(handle)
    #print dict_contacts
    for intf_name in dict_contacts:
        ct = 0
        for ch1_res in dict_contacts[intf_name].keys():
            for ch2_res in dict_contacts[intf_name][ch1_res]:
                ct += 1
        dict_num_contacts[intf_name] = ct
    return dict_num_contacts

def filter_csv(csvfile,outfile):
    '''
    Takes the CSV file as input and does a sanity check that all features are computed
    '''
    if os.path.isfile(outfile):
        os.remove(outfile)
    out = open(outfile,'w')
    with open(csvfile,'r') as infile:
        for lne in infile:
            if 'NA' not in lne:
                out.write(lne)
    out.close() 

def make_predictions(saved_sc, csvfile, saved_model, outfle, session_id):
    with open(saved_model,'rb') as file:
        clf_m = pickle.load(file)

    topredict = pd.read_csv(csvfile)
    if topredict.empty:
        print 'All features could not be calculated successfully for atleast one of the interfaces!'
        print 'Exiting'
        sys.exit(1)
    topredict_1 = topredict.drop(['pdb','interface','Num_intf_residues'], axis=1)
    with open(saved_sc, 'rb') as file:
        sc = pickle.load(file)
        topredict_1 = np.nan_to_num(sc.transform(topredict_1))
    
    pred_em = clf_m.predict(topredict_1)
    dec = clf_m.decision_function(topredict_1)

    out = open(outfle, 'w')
    out.write('#PDB,chains,predicted_class,pi_score\n')


    pdb_em = topredict['pdb']
    pdb_em_int = topredict['interface']

    for i in range(len(topredict_1)):
        out.write(str(pdb_em[i]) + ',' +
                  str(pdb_em_int[i]) + ',' +
                  str(pred_em[i]) + ',' +
                  str(round(dec[i],2)) + '\n')
    out.close()



# -*- coding: utf-8 -*-
#Adapted from sc_utils.py (https://gitlab.com/topf-lab/pi_score/-/tree/master/score_scripts)
import os
import json
import uuid
import subprocess

def write_sc_script(pdbfile,
                    chain_lst=None,
                    outfle=None,
                    ):
    '''
    chain_lst should be list of only two chain Ids e.g. [A,B]
    '''
    #print ('Writing sc script', chain_lst)
    assert (chain_lst != None),"Input chain ID list!"
    assert len(chain_lst) == 2, "Input list with only two chain IDs e.g [A,B]"
    if outfle == None:
        outfle = pdbfile + '_' + '_'.join(chain_lst) + '_sc_infle.sh'
    #print (outfle, 'OUTFLE for SC')
    out = open(outfle,'w')
    out.write('#!/bin/bash\n')
    out.write('sc XYZIN ' + pdbfile + ' <<eof \n')
    out.write('MOLECULE 1\nCHAIN ' + chain_lst[0] + '\n' + 'MOLECULE 2\nCHAIN ' + chain_lst[1] + '\nEND\neof')
    out.close()
    os.system('chmod 777 ' + outfle)
    return outfle

#def run_sc(scriptfile):
#    session_id = uuid.uuid4().hex[:8]


#    unique_script = "%s_%s.sh" % (scriptfile.rsplit('.', 1)[0], session_id)
#    tmp_out = "tmp_sc_%s.out" % session_id
#    os.system("cp %s %s" % (scriptfile, unique_script))

#    os.system("sed -i 's/session name/session_%s/g' %s" % (session_id, unique_script))

#    cmd = "./%s > tmp_sc_%s.out" % (unique_script, session_id)
#    os.system(cmd)
#    parse_sc_output(tmp_out)

def run_sc(scriptfile):
    session_id = uuid.uuid4().hex[:8]
    unique_script = "%s_%s.sh" % (scriptfile.rsplit('.', 1)[0], session_id)
    tmp_out = "%s_tmp_sc_%s.out" % (scriptfile.rsplit('.', 1)[0], session_id)

    subprocess.call(["cp", scriptfile, unique_script])
    subprocess.call(["sed", "-i", "s/session name/session_%s/g" % session_id, unique_script])



    with open(tmp_out, "w") as out_fh:
        proc = subprocess.Popen([unique_script], stdout=out_fh, stderr=subprocess.STDOUT, close_fds=True)
        proc.wait()

    os.remove(unique_script)
    parse_sc_output(tmp_out)

def parse_sc_output(outfle):
    new_name = outfle.rsplit('_tmp_sc.out',1)[0] + '_dict_sc_scores.json'
    if os.path.isfile(new_name):
        with open(new_name) as json_file:
            dict_out = json.load(json_file)
    #else:
    dict_out = {}
    ch_lst = []
    pdb = ''
    sc = ''
    #print(outfle)
    with open(outfle,'r') as out:
        for lne in out:
            if 'Number of atoms selected in chain' in lne:
                ch_lst.append(lne.split('=')[0].split()[-1])
            if 'Logical name: XYZIN  File name: ' in lne:
                #pdb = lne.split()[-1][0:4]
                #for em challenge targets
                pdb = lne.split()[-1].split('_clean')[0]

            if 'Shape complementarity statistic Sc' in lne:
                sc = lne.split('=')[-1].strip()
        try:
            if pdb in dict_out.keys():
                dict_out[pdb].update({'_'.join(ch_lst):sc})
            else:
                dict_out[pdb] = {'_'.join(ch_lst):sc}
        except:
            pass

        with open(new_name,'w') as sec_outfle:
            json.dump(dict_out,sec_outfle)







# HInt_entry.py
import multiprocessing as mp
from absl import app
from HInt.HInt import main as HInt_main

def main():
    mp.set_start_method("spawn", force=True)  # important pour propagation Ctrl+C
    app.run(HInt_main)

if __name__ == "__main__":
    main()
